#include <edapi/journal/journal.h>
#include <edapi/journal/file.h>
#include <edapi/util.h>

#include <gio/gio.h>
#include <stdio.h>

typedef struct {
    gchar *location;
    GFile *loc;
    GList *files;
    GFileMonitor *monitor;
} EDJournalPrivate;

struct _EDJournal {
    GObject parent;
};

struct _EDJournalClass {
    GObjectClass parent_class;
};

G_DEFINE_TYPE_EXTENDED(
    EDJournal,
    ed_journal,
    G_TYPE_OBJECT,
    0,
    G_ADD_PRIVATE(EDJournal)
    );

static void ed_journal_changed(GFileMonitor *self, GFile *file, GFile *other,
                               GFileMonitorEvent event,
                               gpointer user);

static void ed_journal_finalize(GObject *obj)
{
    EDJournal *self = ED_JOURNAL(obj);
    EDJournalPrivate *p = ed_journal_get_instance_private(self);

    free(p->location);
    p->location = NULL;

    G_OBJECT_CLASS(ed_journal_parent_class)->finalize(obj);
}

static void ed_journal_dispose(GObject *obj)
{
    EDJournal *self = ED_JOURNAL(obj);
    EDJournalPrivate *p = ed_journal_get_instance_private(self);

    g_clear_object(&p->loc);
    g_clear_object(&p->monitor);

    G_OBJECT_CLASS(ed_journal_parent_class)->dispose(obj);
}

static void ed_journal_class_init(EDJournalClass *klass)
{
    g_signal_new("new-entry",
                 G_TYPE_FROM_CLASS(klass),
                 G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
                 0, NULL, NULL, NULL,
                 G_TYPE_NONE,
                 2,
                 ED_TYPE_JOURNALFILE,
                 ED_TYPE_JOURNALENTRY
        );

    G_OBJECT_CLASS(klass)->finalize = ed_journal_finalize;
    G_OBJECT_CLASS(klass)->dispose = ed_journal_dispose;
}

gchar *ed_journal_determine_location(void)
{
    gchar *location = NULL;
    gchar *tmp = NULL;

    /**
     * on Windows we have %USERPROFILE% which points to the current
     * users home directory. See if they have a Saved Games folder.
     */
    char const *env1 = getenv("%USERPROFILE%");
    char const *env2 = getenv("USERPROFILE");
    if (!S_EMPTY(env1) || !S_EMPTY(env2)) {
        tmp = g_build_path(
            G_DIR_SEPARATOR_S,
            S_EMPTY(env1) ? env2 : env1,
            "Saved Games",
            "Frontier Developments",
            "Elite Dangerous",
            NULL
            );

        if (g_file_test(tmp, G_FILE_TEST_IS_DIR)) {
            location = tmp;
            tmp = NULL;
        }
    }

    if (S_EMPTY(location)) {
        return NULL;
    }

    if (!g_file_test(location, G_FILE_TEST_IS_DIR)) {
        g_free(location);
        return NULL;
    }

    return location;
}

static void ed_journal_init(EDJournal *self)
{
}

EDJournal *ed_journal_new(void)
{
    return g_object_new(ED_TYPE_JOURNAL, NULL);
}

EDJournal *ed_journal_new_with_location(gchar const *dir, GError **error)
{
    EDJournal *j = NULL;
    EDErrorCode ret = ed_error_success;
    GError *tmp = NULL;

    if (dir == NULL) {
        gchar *loc = ed_journal_determine_location();
        return_if_true(loc == NULL, NULL);

        j = ed_journal_new_with_location(loc, error);
        g_free(loc);

        return j;
    }

    j = ed_journal_new();
    return_if_true(j == NULL, NULL);

    if (!S_EMPTY(dir)) {
        ret = ed_journal_open(j, dir, &tmp);
        if (ED_ERROR(ret)) {
            g_propagate_error(error, tmp);
            g_clear_object(&j);
            return NULL;
        }
    }

    return j;
}

static void ed_journal_load_files(EDJournal *self)
{
    EDJournalPrivate *p = ed_journal_get_instance_private(self);
    GDir *loc = NULL;
    gchar const *name = NULL;
    EDErrorCode ret = ed_error_success;
    EDJournalFile *journalfile = NULL;

    g_list_free_full(p->files, g_object_unref);
    p->files = NULL;

    loc = g_dir_open(p->location, 0, NULL);
    goto_if_true(loc == NULL, done);

    while ((name = g_dir_read_name(loc)) != NULL) {
        if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0) {
            continue;
        }

        gchar *full = g_build_path(
            G_DIR_SEPARATOR_S,
            p->location,
            name,
            NULL
            );

        if (g_file_test(full, G_FILE_TEST_IS_DIR)) {
            goto next;
        }

        ret = ed_journal_file_parse_filename(name, NULL, NULL);
        goto_if_error(ret, next);

        journalfile = ed_journal_file_new();
        goto_if_true(journalfile == NULL, next);

        ret = ed_journal_file_peek(journalfile, full, NULL);
        if (ED_SUCCESS(ret)) {
            p->files = g_list_append(p->files, journalfile);
        }

    next:

        g_free(full);
    }

    p->files = g_list_sort(
        p->files, (GCompareFunc)ed_journal_file_compare_asc);

done:

    if (loc != NULL) {
        g_dir_close(loc);
        loc = NULL;
    }
}

static EDErrorCode ed_journal_watch(EDJournal *self, GError **error)
{
    return_if_true(self == NULL, ed_error_args);

    EDJournalPrivate *p = ed_journal_get_instance_private(self);
    GError *tmp = NULL;

    g_clear_object(&p->monitor);

    p->monitor = g_file_monitor_directory(p->loc, 0, NULL, &tmp);
    if (p->monitor == NULL) {
        g_propagate_error(error, tmp);
        return ed_error_internal;
    }

    g_signal_connect(
        p->monitor, "changed", G_CALLBACK(ed_journal_changed), self);

    return ed_error_success;
}

static void ed_journal_changed(GFileMonitor *mon, GFile *file, GFile *other,
                               GFileMonitorEvent event,
                               gpointer user)
{
    EDJournal *self = ED_JOURNAL(user);

    if (event == G_FILE_MONITOR_EVENT_CHANGED) {
        printf("file changed!\n");
    }

    (void)self;
}

gchar const *ed_journal_get_location(EDJournal *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalPrivate *p = ed_journal_get_instance_private(self);
    return p->location;
}

void ed_journal_close(EDJournal *self)
{
    return_if_true(self == NULL,);

    EDJournalPrivate *p = ed_journal_get_instance_private(self);

    g_free(p->location);
    p->location = NULL;

    g_clear_object(&p->loc);

    g_clear_object(&p->monitor);

    g_list_free_full(p->files, g_object_unref);
    p->files = NULL;
}

EDErrorCode ed_journal_open(EDJournal *self, gchar const *dir, GError **error)
{
    return_if_true(self == NULL || dir == NULL, ed_error_args);

    EDJournalPrivate *p = ed_journal_get_instance_private(self);
    EDErrorCode ret = 0;
    GError *tmp = NULL;

    if (!g_file_test(dir, G_FILE_TEST_IS_DIR)) {
        return ed_error_invalid;
    }

    g_free(p->location);
    p->location = g_strdup(dir);

    g_clear_object(&p->loc);
    p->loc = g_file_new_for_path(dir);

    ed_journal_load_files(self);

    ret = ed_journal_watch(self, &tmp);
    if (ED_ERROR(ret)) {
        g_propagate_error(error, tmp);
        ed_journal_close(self);
        return ed_error_internal;
    }

    return ed_error_success;
}

GList *ed_journal_get_files(EDJournal *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalPrivate *p = ed_journal_get_instance_private(self);
    return p->files;
}
